﻿using Microsoft.Win32;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Drawing;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Security.Cryptography.X509Certificates;
using System.Threading;
using System.Threading.Tasks;
using System.Windows.Forms;
using Titanium.Web.Proxy;
using Titanium.Web.Proxy.EventArguments;
using Titanium.Web.Proxy.Models;

using System.Linq;
using System.Globalization;
using System.Text.RegularExpressions;

namespace Redirect
{
    public class Program
    {
        static Mutex mutex = new Mutex(true, "{D6F595AA-BB26-4DC6-1181-F214210859AE}");

        [DllImport("user32.dll")]
        private static extern bool PostMessage(IntPtr hwnd, uint msg, IntPtr wparam, IntPtr lparam);
        [DllImport("user32.dll")]
        private static extern uint RegisterWindowMessage(string message);
        private const int HWND_BROADCAST = 0xFFFF;
        public static uint WM_REDIRECTCLOSE = RegisterWindowMessage("WM_REDIRECTCLOSE");

        [STAThread]
        public static void Main(string[] args)
        {
            if (args.Length > 0 && args[0].ToLower() == "/close")
            {
                PostMessage((IntPtr)HWND_BROADCAST, WM_REDIRECTCLOSE, IntPtr.Zero, IntPtr.Zero);
                return;
            }
            if (args.Length > 0 && (args[0].ToLower() == "/clear" || args[0].ToLower() == "/reset"))
            {
                //Clear settings
                int error = 0;
                bool success = ProxyMethods.SetProxySettings(false, null, "<-loopback>", null, true, out error);
                if (!success)
                {
                    MessageBox.Show("Could not clear system proxy settings. You will need to change settings manually. Error code: " + error.ToString(), "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
                return;
            }

            if (mutex.WaitOne(TimeSpan.Zero, true))
            {
                Application.EnableVisualStyles();
                Application.SetCompatibleTextRenderingDefault(false);
                Application.Run(new Form1());
            }
        }
    }

    public class Form1 : Form
    {
        private ProxyController controller;

        private TextBox textBox;

        private readonly object @lock = new object();

        private bool ignoreCloseMessage = false;

        private bool alwaysOnTop = false;

        private string version = "4.0";

        public Form1()
        {
            this.Text = "Flashpoint Redirector v" + version;
            this.Size = new Size(600, 400);
            this.WindowState = FormWindowState.Minimized;
            try
            {
                this.Icon = Icon.ExtractAssociatedIcon(Assembly.GetExecutingAssembly().Location);
            }
            catch
            {
            }

            textBox = new TextBox();
            textBox.MaxLength = int.MaxValue;
            textBox.Multiline = true;
            textBox.ReadOnly = true;
            textBox.ScrollBars = ScrollBars.Both;
            textBox.Font = new Font(textBox.Font.FontFamily, 12f);
            textBox.Dock = DockStyle.Fill;
            textBox.BackColor = SystemColors.Window;

            this.Controls.Add(textBox);

            MenuStrip menuStrip = new MenuStrip();
            ToolStripMenuItem toolsMenu = new ToolStripMenuItem("Tools");
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Copy all text", null, (o, e) => { this.CopyAllText(); }));
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Copy all URLs", null, (o, e) => { this.CopyAllUrls(); }));
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Copy all 404 Not Found URLs", null, (o, e) => { this.Copy404NotFoundUrls(); }));
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Copy all PATHCHECK warnings", null, (o, e) => { this.CopyPathcheckWarnings(); }));
            toolsMenu.DropDownItems.Add(new ToolStripSeparator());
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Clear all text", null, (o, e) => { this.ClearText(); }));
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Clear WinINet cache", null, (o, e) => { this.ClearCache(); }));
            toolsMenu.DropDownItems.Add(new ToolStripSeparator());
            toolsMenu.DropDownItems.Add(new ToolStripMenuItem("Always on top", null, (o, e) => { this.AlwaysOnTop(o); }));

            menuStrip.Items.Add(toolsMenu);

            this.Controls.Add(menuStrip);
        }

        protected override void OnLoad(EventArgs e)
        {
            AddText("Flashpoint Redirector v" + version);

            string permissionError = "";

            //Registry permission
            try
            {
                Registry.CurrentUser.OpenSubKey("Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings", true).Close();
            }
            catch (System.Security.SecurityException)
            {
                permissionError = "Could not write to registry. Make sure you have registry write permission.";
            }

            if (permissionError != "")
            {
                this.ShowOnTop();

                string errorText = "Error with permissions: " + permissionError + Environment.NewLine + Environment.NewLine + "Games may not work until the error is fixed. Close Flashpoint and open it again once the error is fixed.";

                textBox.Clear();
                textBox.AppendText(errorText);

                MessageBox.Show(this, errorText, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            else
            {
                try
                {
                    controller = new ProxyController(this);
                    controller.StartProxy();
                }
                catch (Exception ex)
                {
                    this.ShowOnTop();

                    string errorText = "Error with proxy: " + ex.Message + Environment.NewLine + Environment.NewLine + "Games may not work until the error is fixed. Close Flashpoint and open it again once the error is fixed.";

                    textBox.Clear();
                    textBox.AppendText(errorText);

                    MessageBox.Show(this, errorText, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        protected override void WndProc(ref Message m)
        {
            if (m.Msg == Program.WM_REDIRECTCLOSE)
            {
                ignoreCloseMessage = true;
                this.Close();
            }
            else
            {
                base.WndProc(ref m);
            }
        }

        protected override void OnFormClosing(FormClosingEventArgs e)
        {
            if (e.CloseReason == CloseReason.UserClosing && !ignoreCloseMessage)
            {
                if (MessageBox.Show("Are you sure you want to close Flashpoint Redirector?" + Environment.NewLine + Environment.NewLine + "Games may not work if Flashpoint Redirector is closed.", "Flashpoint Redirector", MessageBoxButtons.YesNo, MessageBoxIcon.Warning) == DialogResult.No)
                {
                    e.Cancel = true;
                    return;
                }
            }

            if (controller != null)
            {
                try
                {
                    controller.Stop();
                }
                catch (Exception ex)
                {
                    this.ShowOnTop();

                    string errorText = "Error stopping proxy: " + Environment.NewLine + ex.Message;

                    MessageBox.Show(this, errorText, "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                }
            }
        }

        private void CopyAllText()
        {
            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() =>
                {
                    if (!string.IsNullOrEmpty(textBox.Text)) { Clipboard.SetText(textBox.Text); }
                    else { Clipboard.Clear(); }
                }));
            }

            MessageBox.Show(this, "Copied all text");
        }

        private void CopyAllUrls()
        {
            int amount = 0;

            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() =>
                {
                    string[] lines = textBox.Text.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);

                    string clipboard = "";

                    foreach (string line in lines)
                    {
                        int start = line.ToLower().IndexOf("http://");
                        if (start != -1)
                        {
                            clipboard += line.Substring(start) + Environment.NewLine;
                            amount++;
                        }
                    }

                    if (!string.IsNullOrEmpty(clipboard)) { Clipboard.SetText(clipboard); }
                    else { Clipboard.Clear(); }
                }));
            }

            MessageBox.Show(this, "Copied " + amount + " URLs");
        }

        private void Copy404NotFoundUrls()
        {
            int amount = 0;

            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() =>
                {
                    string[] lines = textBox.Text.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);

                    string clipboard = "";

                    foreach (string line in lines)
                    {
                        if (line.ToLower().StartsWith("(404"))
                        {
                            int start = line.ToLower().IndexOf("http://");
                            if (start != -1)
                            {
                                clipboard += line.Substring(start) + Environment.NewLine;
                                amount++;
                            }
                        }
                    }

                    if (!string.IsNullOrEmpty(clipboard)) { Clipboard.SetText(clipboard); }
                    else { Clipboard.Clear(); }
                }));
            }

            MessageBox.Show(this, "Copied " + amount + " 404 Not Found URLs");
        }

        private void CopyPathcheckWarnings()
        {
            int amount = 0;

            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() =>
                {
                    string[] lines = textBox.Text.Split(new string[] { Environment.NewLine }, StringSplitOptions.RemoveEmptyEntries);

                    string clipboard = "";

                    foreach (string line in lines)
                    {
                        if (line.ToLower().StartsWith("pathcheck:"))
                        {
                            clipboard += line.Substring(11) + Environment.NewLine;
                            amount++;
                        }
                    }

                    if (!string.IsNullOrEmpty(clipboard)) { Clipboard.SetText(clipboard); }
                    else { Clipboard.Clear(); }
                }));
            }

            MessageBox.Show(this, "Copied " + amount + " PATHCHECK warnings");
        }

        private void ClearText()
        {
            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() => { textBox.Clear(); }));
            }
        }

        private void ClearCache()
        {
            Process process = null;
            try
            {
                process = Process.Start("rundll32.exe", "inetcpl.cpl,ClearMyTracksByProcess 8");
            }
            catch
            {
                MessageBox.Show(this, "Could not clear WinINet cache", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }
            finally
            {
                if (process != null) { process.Dispose(); }
            }

            MessageBox.Show(this, "Cleared WinINet cache");
        }

        private void AlwaysOnTop(object sender)
        {
            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() =>
                {
                    ToolStripMenuItem menuItem = (ToolStripMenuItem)sender;

                    menuItem.Checked = !menuItem.Checked;

                    this.alwaysOnTop = menuItem.Checked;

                    this.TopMost = menuItem.Checked;
                }));
            }
        }

        public void AddText(string text)
        {
            lock (@lock)
            {
                this.Invoke(new MethodInvoker(() => { textBox.AppendText(text + Environment.NewLine); textBox.SelectionStart = textBox.Text.Length; }));
            }
        }

        public void ShowOnTop()
        {
            this.Invoke(new MethodInvoker(() =>
            {
                this.WindowState = FormWindowState.Normal;

                this.TopMost = true; this.TopMost = false;

                if (this.alwaysOnTop) this.TopMost = true;
            }));
        }
    }

    public class ProxyController
    {
        private readonly SemaphoreSlim @lock = new SemaphoreSlim(1);

        private ProxyServer proxyServer;

        private ExternalProxy localRedirectProxyHttp = new ExternalProxy() { HostName = "localhost", Port = 22500, BypassLocalhost = false };

        private ExternalProxy originalRedirectProxyHttp = null;

        private ExplicitProxyEndPoint explicitEndPointHttp = new ExplicitProxyEndPoint(IPAddress.Loopback, 8888, false);

        private bool originalAutoDetect = true;

        private string originalAutoConfigUrl = "";

        private bool originalProxyEnable = false;

        private string originalProxyOverride = "";

        private string originalProxyServer = "";

        private bool proxySettingsSet = false;

        private Form1 form;

        private class ProcessRedirectClass
        {
            public string check;
            public string value;
        }

        private List<ProcessRedirectClass> processRedirectList = new List<ProcessRedirectClass>();

        public ProxyController(Form1 form1)
        {
            form = form1;

            //Load settings
            if (File.Exists("settings.txt"))
            {
                string[] lines = File.ReadAllLines("settings.txt");

                foreach (string line in lines)
                {
                    if (line.Trim() == "") continue;

                    if (!line.StartsWith("#"))
                    {
                        string[] split = line.Split(',');

                        if (split.Length < 2) continue;

                        processRedirectList.Add(new ProcessRedirectClass() { check = split[0], value = split[1] });
                    }
                }
            }
        }

        private Task<ExternalProxy> GetUpstreamLocalProxy(SessionEventArgsBase session)
        {
            Process process = null;
            try
            {
                process = Process.GetProcessById(session.HttpClient.ProcessId.Value);
            }
            catch
            {
                //Process does not exist
                return Task.FromResult<ExternalProxy>(originalRedirectProxyHttp);
            }

            foreach (ProcessRedirectClass proc in processRedirectList)
            {
                bool matches = false;

                try
                {
                    if (proc.check == "ProcessNameStartsWith") { matches = process.ProcessName.ToLower().StartsWith(proc.value.ToLower()); }
                    else if (proc.check == "ProcessNameContains") { matches = process.ProcessName.ToLower().Contains(proc.value.ToLower()); }
                    else if (proc.check == "ProcessNameEndsWith") { matches = process.ProcessName.ToLower().EndsWith(proc.value.ToLower()); }
                    else if (proc.check == "ProcessFilePathStartsWith") { matches = process.MainModule.FileName.ToLower().StartsWith(proc.value.ToLower()); }
                    else if (proc.check == "ProcessFilePathContains") { matches = process.MainModule.FileName.ToLower().Contains(proc.value.ToLower()); }
                    else if (proc.check == "ProcessFilePathEndsWith") { matches = process.MainModule.FileName.ToLower().EndsWith(proc.value.ToLower()); }
                }
                catch
                {
                }

                if (matches)
                {
                    return Task.FromResult(localRedirectProxyHttp);
                }
            }

            return Task.FromResult<ExternalProxy>(originalRedirectProxyHttp);
        }

        private async void ExceptionFunc(Exception exception)
        {
            await @lock.WaitAsync();

            try
            {
                if (exception is Titanium.Web.Proxy.Exceptions.ProxyHttpException)
                {
                    Titanium.Web.Proxy.Exceptions.ProxyHttpException phex = exception as Titanium.Web.Proxy.Exceptions.ProxyHttpException;
                    Exception innerEx = phex.InnerException;

                    if (innerEx.Message.ToLower().StartsWith("could not establish connection to") && innerEx.InnerException != null && innerEx.InnerException is SocketException)
                    {
                        innerEx = innerEx.InnerException;
                    }

                    //Ignore processes that are not redirected
                    ExternalProxy eProxy = await GetUpstreamLocalProxy(phex.SessionEventArgs);
                    if (eProxy == null)
                    {
                        return;
                    }

                    int errorCode = -1;
                    if (innerEx != null && innerEx is SocketException)
                    {
                        SocketException socketException = innerEx as SocketException;
                        errorCode = socketException.ErrorCode;
                    }

                    //Web server
                    if (innerEx != null && errorCode == 10061)
                    {
                        form.AddText(Environment.NewLine + "Error when redirecting. Make sure the web server is running and is not blocked by a firewall. Error: " + Environment.NewLine + exception.Message + ((innerEx != null) ? ": " + innerEx.Message : "") + Environment.NewLine);
                    }

                    //Firewall
                    else if (innerEx != null && errorCode == 10013)
                    {
                        form.AddText(Environment.NewLine + "Error when redirecting. Make sure a firewall is not blocking Flashpoint Redirector. Error: " + Environment.NewLine + exception.Message + ((innerEx != null) ? ": " + innerEx.Message : "") + Environment.NewLine);
                    }

                    //Ignore errors when changing games
                    else if (innerEx != null && innerEx is Titanium.Web.Proxy.Exceptions.ServerConnectionException)
                    {
                    }

                    else if (innerEx != null && innerEx is IOException)
                    {
                        IOException ioe = innerEx as IOException;
                        if (ioe.HResult == -2146232800 || ioe.Message.ToLower().Contains("was aborted by the software"))
                        {
                        }
                        else
                        {
                            form.AddText(Environment.NewLine + "Error when redirecting. IO error. Error: " + Environment.NewLine + exception.Message + ": " + ioe.Message + Environment.NewLine);
                        }
                    }

                    //Other
                    else
                    {
                        form.AddText(Environment.NewLine + "Error when redirecting. General error. Error: " + Environment.NewLine + exception.Message + ((innerEx != null) ? ": " + innerEx.Message : "") + Environment.NewLine);
                    }
                }
                else
                {
                    Exception innerEx = exception.InnerException;

                    if (innerEx != null && innerEx is SocketException)
                    {
                        SocketException socketException = innerEx as SocketException;
                        //Ignore host not found errors
                        if (socketException.ErrorCode == 11001)
                        {
                        }
                        else
                        {
                            form.AddText(Environment.NewLine + "Socket error: " + exception.Message + ((innerEx != null) ? ": " + innerEx.Message : "") + Environment.NewLine);
                        }
                    }
                    else
                    {
                        form.AddText(Environment.NewLine + "Error: " + exception.Message + ((innerEx != null) ? ": " + innerEx.Message : "") + Environment.NewLine);
                    }
                }
            }
            finally
            {
                @lock.Release();
            }
        }

        public void StartProxy()
        {
            if (Type.GetType("Mono.Runtime") != null)
            {
                throw new Exception("Mono Runtime does not support system proxy settings");
            }

            proxyServer = new ProxyServer(false);

            proxyServer.BufferSize = 32768;
            proxyServer.EnableConnectionPool = true;
            proxyServer.ConnectionTimeOutSeconds = 5;
            proxyServer.MaxCachedConnections = 100;

            proxyServer.CertificateManager.SaveFakeCertificates = false;
            //Do not create certificate file
            proxyServer.CertificateManager.RootCertificate = new X509Certificate2();

            proxyServer.ForwardToUpstreamGateway = false;

            proxyServer.GetCustomUpStreamProxyFunc = GetUpstreamLocalProxy;

            proxyServer.BeforeResponse += BeforeResponse;

            proxyServer.ExceptionFunc = ExceptionFunc;

            try
            {
                proxyServer.AddEndPoint(explicitEndPointHttp);
            }
            catch (Exception ex)
            {
                throw new Exception("Error adding local proxy HTTP endpoint \"" + explicitEndPointHttp.IpAddress + ":" + explicitEndPointHttp.Port + "\": " + ex.Message);
            }

            bool success = false;
            int error = 0;

            //Store original settings
            success = ProxyMethods.GetProxySettings(out originalProxyEnable, out originalProxyServer, out originalProxyOverride, out originalAutoConfigUrl, out originalAutoDetect, out error);
            if (!success)
            {
                throw new Exception("Error reading original system proxy settings. Make sure you have registry read permission. Error code: " + error.ToString());
            }

            //Closed incorrectly
            string oldHttp = "http=localhost:" + explicitEndPointHttp.Port;
            if (originalProxyServer != null && originalProxyServer.Contains(oldHttp))
            {
                originalProxyServer = originalProxyServer.Replace(";" + oldHttp, "");
                originalProxyServer = originalProxyServer.Replace(oldHttp + ";", "");
                originalProxyServer = originalProxyServer.Replace(oldHttp, "");
            }

            //Add default port if not set
            bool defaultPortSet = false;
            string defaultPortOriginalProxyServer = originalProxyServer;
            if (defaultPortOriginalProxyServer != null)
            {
                string[] lines = defaultPortOriginalProxyServer.Split(';');

                for (int i = 0; i < lines.Length; i++)
                {
                    if (lines[i].IndexOf(':') == -1)
                    {
                        defaultPortSet = true;
                        if (lines[i].StartsWith("http=")) { lines[i] += ":80"; }
                        else if (lines[i].StartsWith("https=")) { lines[i] += ":443"; }
                        else if (lines[i].StartsWith("ftp=")) { lines[i] += ":21";  }
                        else if (lines[i].StartsWith("socks=")) { lines[i] += ":1080"; }
                    }
                }

                defaultPortOriginalProxyServer = String.Join(";", lines);
            }

            if (defaultPortSet)
            {
                success = ProxyMethods.SetProxySettings(originalProxyEnable, defaultPortOriginalProxyServer, originalProxyOverride, originalAutoConfigUrl, originalAutoDetect, out error);
                if (!success)
                {
                    throw new Exception("Error writing original system proxy settings. Make sure you have registry write permission. Error code: " + error.ToString());
                }
                proxySettingsSet = true;

                Thread.Sleep(1000);
            }

            try
            {
                proxyServer.Start();
            }
            catch (Exception ex)
            {
                if (ex.InnerException != null && ex.InnerException is SocketException)
                {
                    SocketException socketException = ex.InnerException as SocketException;
                    if (socketException.ErrorCode == 10048)
                    {
                        throw new Exception("HTTP redirection port " + explicitEndPointHttp.Port + " is already in use. Make sure an existing application is not using the port.");
                    }
                }
                throw new Exception("Error starting proxy server: " + ex.Message);
            }

            form.AddText("HTTP redirection started on port " + explicitEndPointHttp.Port);

            form.AddText("Redirecting HTTP traffic to " + explicitEndPointHttp.IpAddress + ":" + localRedirectProxyHttp.Port + " if:");

            if (processRedirectList.Count == 0)
            {
                form.AddText("\tERROR: No redirection rules found");

                form.ShowOnTop();
            }
            else
            {
                foreach (ProcessRedirectClass proc in processRedirectList)
                {
                    form.AddText("\t" + proc.check + " " + proc.value);
                }
            }

            form.AddText("All HTTP URLs below have been redirected to " + explicitEndPointHttp.IpAddress + ":" + localRedirectProxyHttp.Port + Environment.NewLine);

            //Keep original proxy server settings
            //Replace HTTP proxy server with local proxy server
            string newProxyServer = "http=localhost:" + explicitEndPointHttp.Port;

            if (originalProxyServer != null)
            {
                string[] lines = originalProxyServer.Split(';');

                foreach (string line in lines)
                {
                    if (!line.StartsWith("http="))
                    {
                        newProxyServer += ";" + line;
                    }
                    else
                    {
                        if (line == "http=localhost:" + explicitEndPointHttp.Port || line == "http=127.0.0.1:" + explicitEndPointHttp.Port)
                        {
                            originalRedirectProxyHttp = null;
                        }
                        else
                        {
                            string hostnamePortLine = line;
                            if (hostnamePortLine.IndexOf(':') == -1)
                            {
                                hostnamePortLine += ":80";
                            }
                            string[] hostnamePort = hostnamePortLine.Replace("http=", "").Split(':');
                            originalRedirectProxyHttp = new ExternalProxy() { HostName = hostnamePort[0], Port = int.Parse(hostnamePort[1]), BypassLocalhost = false };
                        }
                    }
                }
            }

            //Set proxy server
            //Remove bypass local proxy and exceptions
            //Remove automatic configuration URL
            //Set auto detect
            success = ProxyMethods.SetProxySettings(true, newProxyServer, "<-loopback>", "", false, out error);
            if (!success)
            {
                throw new Exception("Error writing new system proxy settings. Make sure you have registry write permission. Error code: " + error.ToString());
            }
            proxySettingsSet = true;

            Thread.Sleep(1000);

            //Confirm proxy
            bool isProxySet = false;

            bool proxyEnable = false;
            string autoConfigUrl = null;
            string proxyOverride = null;
            string proxyString = null;
            bool autoDetect = true;

            success = ProxyMethods.GetProxySettings(out proxyEnable, out proxyString, out proxyOverride, out autoConfigUrl, out autoDetect, out error);
            if (!success)
            {
                throw new Exception("Error confirming new system proxy settings. Make sure you have registry read permission. Error code: " + error.ToString());
            }

            if (proxyEnable && autoConfigUrl == null && (proxyOverride != null && proxyOverride == "<-loopback>") && (proxyString != null && proxyString.Contains("localhost:" + explicitEndPointHttp.Port)) && !autoDetect)
            {
                isProxySet = true;
            }

            if (!isProxySet)
            {
                throw new Exception("New system proxy settings were not written to registry. Try closing Flashpoint and opening it again.");
            }
        }

        public void Stop()
        {
            proxyServer.BeforeResponse -= BeforeResponse;

            try
            {
                if (proxyServer.ProxyRunning)
                {
                    proxyServer.Stop();
                }
            }
            catch (Exception ex)
            {
                throw new Exception("Error stopping proxy server. You may need to reset the system proxy settings. Error: " + Environment.NewLine + ex.Message);
            }

            if (proxySettingsSet)
            {
                //Restore original settings
                int error = 0;
                bool success = ProxyMethods.SetProxySettings(originalProxyEnable, originalProxyServer, originalProxyOverride, originalAutoConfigUrl, originalAutoDetect, out error);
                if (!success)
                {
                    throw new Exception("Error restoring original system proxy settings. Make sure you have registry write permission. You may need to reset the system proxy settings. Error code: " + error.ToString());
                }
            }
        }

        private async Task BeforeResponse(object sender, SessionEventArgs e)
        {
            //Only processes that are redirected
            if (e.CustomUpStreamProxyUsed != null && e.CustomUpStreamProxyUsed == localRedirectProxyHttp)
            {
                await WriteToForm("(" + e.HttpClient.Response.StatusCode + " " + e.HttpClient.Response.StatusDescription + ") " + e.HttpClient.Request.Method + " " + e.HttpClient.Request.RequestUri);
                
                // Dri0m comes with a hacky solution to a non-problem on any OS except Windows
                // Verify that path+file is capitalized correctly
                string uri = "" + e.HttpClient.Request.RequestUri;
                uri = uri.Substring(uri.IndexOf("//") + 2);

                var index = uri.LastIndexOf('?');
                if (index != -1)
                {
                    uri = uri.Remove(index);
                }

                if (uri[uri.Length - 1] == '/')
                {
                    uri = uri.Remove(uri.Length - 1);
                }

                var regex = new Regex(@":\d+");
                uri = regex.Replace(uri, "", 1);;

                string exactPath;
                string path = @".\..\..\Server\htdocs\" + uri; // The relative position of redirector and htdocs probably won't ever change again, so this is hopefully fine

                if (e.HttpClient.Response.StatusCode == 200)
                {
                    TryGetExactPath(path, out exactPath);

                    if (exactPath != null && !exactPath.Replace("\\", "/").Contains(uri))
                    {
                        await WriteToForm("PATHCHECK: The requested file `" + uri + "` is not capitalized correctly. Please rename the file (if you're adding a new game) or report this as a bug. Exact path on the disk: `" + exactPath + "`");
                    }
                }
            }
        }

        /// <summary>
        /// Gets the exact case used on the file system for an existing file or directory. https://stackoverflow.com/a/29578292
        /// </summary>
        /// <param name="path">A relative or absolute path.</param>
        /// <param name="exactPath">The full path using the correct case if the path exists.  Otherwise, null.</param>
        /// <returns>True if the exact path was found.  False otherwise.</returns>
        /// <remarks>
        /// This supports drive-lettered paths and UNC paths, but a UNC root
        /// will be returned in title case (e.g., \\Server\Share).
        /// </remarks>
        private static bool TryGetExactPath(string path, out string exactPath)
        {
            bool result = false;
            exactPath = null;

            // DirectoryInfo accepts either a file path or a directory path, and most of its properties work for either.
            // However, its Exists property only works for a directory path.
            DirectoryInfo directory = new DirectoryInfo(path);
            if (File.Exists(path) || directory.Exists)
            {
                List<string> parts = new List<string>();

                DirectoryInfo parentDirectory = directory.Parent;
                while (parentDirectory != null)
                {
                    FileSystemInfo entry = parentDirectory.EnumerateFileSystemInfos(directory.Name).First();
                    parts.Add(entry.Name);

                    directory = parentDirectory;
                    parentDirectory = directory.Parent;
                }

                // Handle the root part (i.e., drive letter or UNC \\server\share).
                string root = directory.FullName;
                if (root.Contains(':'))
                {
                    root = root.ToUpper();
                }
                else
                {
                    string[] rootParts = root.Split('\\');
                    root = string.Join("\\", rootParts.Select(part => CultureInfo.CurrentCulture.TextInfo.ToTitleCase(part)));
                }

                parts.Add(root);
                parts.Reverse();
                exactPath = Path.Combine(parts.ToArray());
                result = true;
            }

            return result;
        }

        private async Task WriteToForm(string message)
        {
            await @lock.WaitAsync();

            try
            {
                form.AddText(message);
            }
            finally
            {
                @lock.Release();
            }
        }
    }

    public class ProxyMethods
    {
        public static bool SetProxySettings(bool proxyEnable, String proxyServer, String proxyOverride, String autoConfigUrl, bool autoDetect, out int error)
        {
            bool success = false;
            error = 0;

            InternetPerConnOption[] options = new InternetPerConnOption[4];

            options[0] = new InternetPerConnOption();
            options[1] = new InternetPerConnOption();
            options[2] = new InternetPerConnOption();
            options[3] = new InternetPerConnOption();
            options[0].dwOption = INTERNET_PER_CONN_FLAGS;
            options[1].dwOption = INTERNET_PER_CONN_PROXY_SERVER;
            options[2].dwOption = INTERNET_PER_CONN_PROXY_BYPASS;
            options[3].dwOption = INTERNET_PER_CONN_AUTOCONFIG_URL;

            int perConnFlags = PROXY_TYPE_DIRECT;

            if (proxyEnable || !String.IsNullOrEmpty(proxyServer))
            {
                perConnFlags |= PROXY_TYPE_PROXY;
            }
            if (autoDetect)
            {
                perConnFlags |= PROXY_TYPE_AUTO_DETECT;
            }
            if (!String.IsNullOrEmpty(autoConfigUrl))
            {
                perConnFlags |= PROXY_TYPE_AUTO_PROXY_URL;
            }

            //Set values
            options[0].value.dwValue = perConnFlags;
            options[1].value.pszValue = IntPtr.Zero;
            options[2].value.pszValue = IntPtr.Zero;
            options[3].value.pszValue = IntPtr.Zero;
            if (!String.IsNullOrEmpty(proxyServer))
            {
                options[1].value.pszValue = Marshal.StringToCoTaskMemAnsi(proxyServer);
            }
            if (!String.IsNullOrEmpty(proxyOverride))
            {
                options[2].value.pszValue = Marshal.StringToCoTaskMemAnsi(proxyOverride);
            }
            if (!String.IsNullOrEmpty(autoConfigUrl))
            {
                options[3].value.pszValue = Marshal.StringToCoTaskMemAnsi(autoConfigUrl);
            }

            //Store options
            IntPtr optionsBufferPtr = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(InternetPerConnOption)) * options.Length);
            IntPtr currentPtr = optionsBufferPtr;
            for (int i = 0; i < options.Length; i++)
            {
                Marshal.StructureToPtr(options[i], currentPtr, false);
                currentPtr = (IntPtr)((long)(currentPtr.ToInt64() + Marshal.SizeOf(options[i])));
            }

            InternetPerConnOptionList optionList = new InternetPerConnOptionList();
            optionList.pszConnection = IntPtr.Zero;
            optionList.dwOptionCount = options.Length;
            optionList.pOptions = optionsBufferPtr;
            optionList.dwSize = Marshal.SizeOf(typeof(InternetPerConnOptionList));
            optionList.dwOptionError = 0;

            int optionListSize = optionList.dwSize;

            //Set proxy settings
            if (InternetSetOption(IntPtr.Zero, INTERNET_OPTION_PER_CONNECTION_OPTION, ref optionList, optionListSize))
            {
                InternetSetOption(IntPtr.Zero, INTERNET_OPTION_PROXY_SETTINGS_CHANGED, IntPtr.Zero, 0);
                success = true;
                error = 0;
            }
            else
            {
                success = false;
                error = Marshal.GetLastWin32Error();
            }

            //Free memory
            Marshal.FreeCoTaskMem(optionsBufferPtr);

            if (options[1].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[1].value.pszValue);
            }
            if (options[2].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[2].value.pszValue);
            }
            if (options[3].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[3].value.pszValue);
            }

            return success;
        }

        public static bool GetProxySettings(out bool proxyEnable, out String proxyServer, out String proxyOverride, out String autoConfigUrl, out bool autoDetect, out int error)
        {
            bool success = false;

            proxyServer = null;
            proxyOverride = null;
            autoConfigUrl = null;
            autoDetect = true;
            proxyEnable = false;
            error = 0;

            autoDetect = AutoDetectEnabled();

            InternetPerConnOption[] options = new InternetPerConnOption[4];

            options[0] = new InternetPerConnOption();
            options[1] = new InternetPerConnOption();
            options[2] = new InternetPerConnOption();
            options[3] = new InternetPerConnOption();
            options[0].dwOption = INTERNET_PER_CONN_FLAGS;
            options[1].dwOption = INTERNET_PER_CONN_PROXY_SERVER;
            options[2].dwOption = INTERNET_PER_CONN_PROXY_BYPASS;
            options[3].dwOption = INTERNET_PER_CONN_AUTOCONFIG_URL;

            //Store options
            IntPtr optionsBufferPtr = Marshal.AllocCoTaskMem(Marshal.SizeOf(typeof(InternetPerConnOption)) * options.Length);
            IntPtr currentPtr = optionsBufferPtr;
            for (int i = 0; i < options.Length; i++)
            {
                Marshal.StructureToPtr(options[i], currentPtr, false);
                currentPtr = (IntPtr)((long)(currentPtr.ToInt64() + Marshal.SizeOf(options[i])));
            }

            InternetPerConnOptionList optionList = new InternetPerConnOptionList();
            optionList.dwSize = Marshal.SizeOf(typeof(InternetPerConnOptionList));
            optionList.pszConnection = IntPtr.Zero;
            optionList.dwOptionCount = options.Length;
            optionList.dwOptionError = 0;
            optionList.pOptions = optionsBufferPtr;

            int optionListSize = optionList.dwSize;

            //Get proxy settings
            if (!InternetQueryOption(IntPtr.Zero, INTERNET_OPTION_PER_CONNECTION_OPTION, ref optionList, ref optionListSize))
            {
                success = false;
                error = Marshal.GetLastWin32Error();
            }
            else
            {
                //Get stored options
                currentPtr = optionsBufferPtr;
                for (int i = 0; i < options.Length; i++)
                {
                    options[i] = (InternetPerConnOption)Marshal.PtrToStructure(currentPtr, typeof(InternetPerConnOption));
                    currentPtr = (IntPtr)((long)(currentPtr.ToInt64() + Marshal.SizeOf(options[i])));
                }

                //Get values
                int perConnFlags = options[0].value.dwValue;

                proxyEnable = ((perConnFlags & PROXY_TYPE_PROXY) == PROXY_TYPE_PROXY);

                if (options[1].value.pszValue != IntPtr.Zero)
                {
                    proxyServer = Marshal.PtrToStringAnsi(options[1].value.pszValue);
                }
                if (options[2].value.pszValue != IntPtr.Zero)
                {
                    proxyOverride = Marshal.PtrToStringAnsi(options[2].value.pszValue);
                }
                if (options[3].value.pszValue != IntPtr.Zero)
                {
                    autoConfigUrl = Marshal.PtrToStringAnsi(options[3].value.pszValue);
                }

                success = true;
                error = 0;
            }

            //Free memory
            Marshal.FreeCoTaskMem(optionsBufferPtr);

            if (options[1].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[1].value.pszValue);
            }
            if (options[2].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[2].value.pszValue);
            }
            if (options[3].value.pszValue != IntPtr.Zero)
            {
                Marshal.FreeHGlobal(options[3].value.pszValue);
            }

            return success;
        }

        private static bool AutoDetectEnabled()
        {
            bool autoDetect = true;
            try
            {
                using (RegistryKey reg = Registry.CurrentUser.OpenSubKey("Software\\Microsoft\\Windows\\CurrentVersion\\Internet Settings\\Connections", false))
                {
                    if (reg != null)
                    {
                        byte[] settings = (reg.GetValue("DefaultConnectionSettings") != null) ? reg.GetValue("DefaultConnectionSettings") as byte[] : null;

                        if (settings != null && settings.Length > 8)
                        {
                            if ((settings[8] & PROXY_TYPE_AUTO_DETECT) == PROXY_TYPE_AUTO_DETECT)
                            {
                                autoDetect = true;
                            }
                            else
                            {
                                autoDetect = false;
                            }
                        }
                    }
                }
            }
            catch
            {
                autoDetect = true;
            }
            return autoDetect;
        }

        private const int INTERNET_OPTION_PER_CONNECTION_OPTION = 75;
        private const int INTERNET_OPTION_PROXY_SETTINGS_CHANGED = 95;

        private const int INTERNET_PER_CONN_FLAGS = 1;
        private const int INTERNET_PER_CONN_PROXY_SERVER = 2;
        private const int INTERNET_PER_CONN_PROXY_BYPASS = 3;
        private const int INTERNET_PER_CONN_AUTOCONFIG_URL = 4;

        private const int PROXY_TYPE_DIRECT = 0x00000001;
        private const int PROXY_TYPE_PROXY = 0x00000002;
        private const int PROXY_TYPE_AUTO_PROXY_URL = 0x00000004;
        private const int PROXY_TYPE_AUTO_DETECT = 0x00000008;

        [StructLayout(LayoutKind.Sequential)]
        private struct InternetPerConnOptionList
        {
            public int dwSize;
            public IntPtr pszConnection;
            public int dwOptionCount;
            public int dwOptionError;
            public IntPtr pOptions;
        }

        [StructLayout(LayoutKind.Sequential)]
        private struct InternetPerConnOption
        {
            public int dwOption;
            public InternetPerConnOptionValue value;
        }

        [StructLayout(LayoutKind.Explicit)]
        private struct InternetPerConnOptionValue
        {
            [FieldOffset(0)]
            public int dwValue;

            [FieldOffset(0)]
            public IntPtr pszValue;

            [FieldOffset(0)]
            public System.Runtime.InteropServices.ComTypes.FILETIME ftValue;
        }

        [DllImport("wininet.dll", SetLastError = true)]
        private static extern bool InternetSetOption(IntPtr hInternet, int dwOption, IntPtr lpBuffer, int dwBufferLength);
        [DllImport("wininet.dll", SetLastError = true)]
        private static extern bool InternetSetOption(IntPtr hInternet, int dwOption, ref InternetPerConnOptionList lpOptionList, int dwBufferLength);
        [DllImport("wininet.dll", SetLastError = true)]
        private static extern bool InternetQueryOption(IntPtr hInternet, int dwOption, ref InternetPerConnOptionList lpOptionList, ref int lpdwBufferLength);
    }
}